(ns core.demo
  (:require [datascript.core :as d]))

;; schema is not required you can use {} but it allows you to create relations and unique attributes in your database
;; some info on schema options here https://docs.datomic.com/on-prem/schema.html
(def schema
  "Database schema layout, applies constraints to certain fields"
  ;; only want the username to exist once
  {:user/name {:db/unique :db.unique/identity}
  ;;our user can belong to many rooms
   :user/rooms {:db/cardinality :db.cardinality/many
                :db/valueType :db.type/ref}
   :room/name {:db/unique :db.unique/identity}})

;; connection to database


(def conn (d/create-conn schema))


;; notice how users room is a vector of maps, if :room/name does not exist its created
;; if it does exists it links via the entity ids, so in this example oly's :user/rooms stores the ids #{7 8}


(def initial-data-payload
  [{:user/name "Oly" :user/rooms [{:room/name "room1"} {:room/name "room2"}]}
   {:user/name "Sam" :user/rooms [:room/name "room1"]}
   {:user/name "Brett" :user/rooms [:room/name "room1"]}
   {:user/name "Cameron"}
   {:user/name "Matt"}
   {:user/name "Zoran"}
   {:room/name "room1" :room/link "room1-link" :room/description "room description"}
   {:room/name "room2" :room/link "room2-link" :room/description "room description"}
   {:room/name "room3" :room/link "room3-link" :room/description "room description"}])

(d/transact! conn initial-data-payload)

;; dealing with entities these are basically unique identifies like ids or primary keys from sql
;; you can look up entities by id or by a unique attribute set in the schema
;; entities return a single item like in a hash map or dictionary
(d/entity @conn 1)
(d/entity @conn [:user/name "Oly"])

;; often we want a lot more details this is where pull comes into play
;; '[*] is a shorthand to pull anything, including relations we can also specify the attributes
(d/pull @conn '[*] 1)
(d/pull @conn '[:user/name :user/rooms] 1)

;; to query data from the database you use d/q and a few notations
;; in the where notice everything is a triple ie an [entity attribute value]
;; if we don't care about something in the eav we use an _ to ignnore
;; we can create variables with ?var-name on any of the triplet


;; find all rooms and there unique entity ids, try doing the same for user perhaps
(d/q '[:find  ?room-name2
       :where [_ :room/name ?room-name2]] @conn)


;; in this example we say find all entities containing the value "room1" in the :user/room attribute
;; bind the entity to a value called ?user-entity
;; in the second where clause we use the ?user-entity which now has a value to find any attributes with :user/name matching the same entity id and bind the result to ?users variable
;; the find then pulls the result from ?users we could add in ?user-entity if we needed it or any other attribute wee wish to pull


(d/q '[:find ?users :where
       [?user-entity :user/rooms "room1"]
       [?user-entity :user/name ?users]] @conn)

;; find room1 entity id, then find all users entities joined to the room entity then find the usernames of the users joined to the room
(d/q '[:find ?room-entity ?user-entity ?user :where
       [?room-entity :room/name "room1"]
       [?user-entity :user/rooms ?room-entity]
       [?user-entity :user/name ?user]] @conn)

;; we could ignore the join data and get a list of all users combined to room1
(d/q '[:find ?room-entity ?room-name ?user :where
       [?room-entity :room/name "room1"]
       [?room-entity :room/name ?room-name]
       [?user-entity :user/name ?user]] @conn)


;; we can use pull inside a query, this will find room1 and pull all data with the same entity
(d/q '[:find (pull ?room-entity [*]) :where
       [?room-entity :room/name "room1"]] @conn)


;; aggregates, how many rooms do we have "." is short hand so we dont get a nested Vector
;; try with out


(d/q '[:find (count ?room-entity) .  :where
       [?room-entity :room/name _]] @conn)

;; we could so a sum instead, summing the entities is a bit silly but you get the idea
(d/q '[:find (sum ?room-entity) . :where
       [?room-entity :room/name _]] @conn)


